"""GeminiAug class for query augmentation using Google's Gemini Model."""

from __future__ import annotations

import google.generativeai as genai

from typeguard import typechecked

from rago.augmented.base import AugmentedBase


@typechecked
class GeminiAug(AugmentedBase):
    """GeminiAug class for query augmentation using Gemini API."""

    default_model_name: str = 'gemini-1.5-flash'
    default_top_k: int = 1

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        genai.configure(api_key=self.api_key)

    def search(
        self, query: str, documents: list[str], top_k: int = 0
    ) -> list[str]:
        """Augment the query by expanding or rephrasing it using Gemini."""
        top_k = top_k or self.top_k
        prompt = self.prompt_template.format(
            query=query, context=' '.join(documents), top_k=top_k
        )

        response = genai.GenerativeModel(self.model_name).generate_content(
            prompt
        )

        augmented_query = str(
            response.text.strip()
            if hasattr(response, 'text')
            else response[0].text.strip()
        )
        return augmented_query.split(self.result_separator)[:top_k]



"""Base classes for generation."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any

import torch

from typeguard import typechecked


@typechecked
class GenerationBase:
    """Generic Generation class."""

    api_key: str = ''
    device_name: str = 'cpu'
    device: torch.device
    model: Any
    model_name: str = ''
    tokenizer: Any
    temperature: float = 0.5
    output_max_length: int = 500
    prompt_template: str = (
        'question: \n```\n{query}\n```\ncontext: ```\n{context}\n```'
    )

    # default parameters that can be overwritten by the derived class
    default_device_name: str = 'cpu'
    default_model_name: str = ''
    default_temperature: float = 0.5
    default_output_max_length: int = 500
    default_prompt_template: str = (
        'question: \n```\n{query}\n```\ncontext: ```\n{context}\n```'
    )

    def __init__(
        self,
        model_name: str = '',
        api_key: str = '',
        temperature: float = 0.5,
        prompt_template: str = '',
        output_max_length: int = 500,
        device: str = 'auto',
    ) -> None:
        """Initialize Generation class.

        Parameters
        ----------
        model_name : str
            The name of the model to use.
        api_key : str
        temperature : float
        prompt_template: str
        output_max_length : int
            Maximum length of the generated output.
        device: str (default=auto)
        """
        self.api_key = api_key
        self.model_name = model_name or self.default_model_name
        self.output_max_length = (
            output_max_length or self.default_output_max_length
        )
        self.temperature = temperature or self.default_temperature

        self.prompt_template = prompt_template or self.default_prompt_template

        if self.device_name not in ['cpu', 'cuda', 'auto']:
            raise Exception(
                f'Device {self.device_name} not supported. '
                'Options: cpu, cuda, auto.'
            )

        cuda_available = torch.cuda.is_available()
        self.device_name = (
            'cpu' if device == 'cpu' or not cuda_available else 'cuda'
        )
        self.device = torch.device(self.device_name)

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        return

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        return

    @abstractmethod
    def generate(
        self,
        query: str,
        context: list[str],
    ) -> str:
        """Generate text with optional language parameter.

        Parameters
        ----------
        query : str
            The input query or prompt.
        context : list[str]
            Additional context information for the generation.

        Returns
        -------
        str
            Generated text based on query and context.
        """
        ...



"""Base classes for the augmented step."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Optional

from typeguard import typechecked

from rago.db import DBBase, FaissDB

if TYPE_CHECKING:
    import numpy as np
    import numpy.typing as npt

    from torch import Tensor


@typechecked
class AugmentedBase:
    """Define the base structure for Augmented classes."""

    api_key: str = ''
    model: Optional[Any]
    model_name: str = ''
    db: Any
    top_k: int = 0
    temperature: float = 0.5
    prompt_template: str = ''
    result_separator = '\n'
    output_max_length: int = 500

    # default values to be overwritten by the derived classes
    default_model_name: str = ''
    default_top_k: int = 0
    default_temperature: float = 0.5
    default_prompt_template: str = (
        'Retrieve {top_k} entries from the context that better answer the '
        'following query:\n```\n{query}\n```\n\ncontext:\n```\n{context}\n```'
    )
    default_result_separator = '\n'
    default_output_max_length: int = 500

    def __init__(
        self,
        model_name: str = '',
        api_key: str = '',
        db: DBBase = FaissDB(),
        top_k: int = 0,
        temperature: float = 0.5,
        prompt_template: str = '',
        result_separator: str = '\n',
        output_max_length: int = 500,
    ) -> None:
        """Initialize AugmentedBase."""
        self.db = db
        self.api_key = api_key

        self.top_k = top_k or self.default_top_k
        self.model_name = model_name or self.default_model_name
        self.temperature = temperature or self.default_temperature
        self.result_separator = (
            result_separator or self.default_result_separator
        )
        self.prompt_template = prompt_template or self.default_prompt_template
        self.output_max_length = (
            output_max_length or self.default_output_max_length
        )

        self.model = None

        self._validate()
        self._setup()

    def _validate(self) -> None:
        """Raise an error if the initial parameters are not valid."""
        return

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        return

    def get_embedding(
        self, content: list[str]
    ) -> list[Tensor] | npt.NDArray[np.float64] | Tensor:
        """Retrieve the embedding for a given text using OpenAI API."""
        raise Exception('Method not implemented.')

    @abstractmethod
    def search(
        self,
        query: str,
        documents: Any,
        top_k: int = 0,
    ) -> list[str]:
        """Search an encoded query into vector database."""
        ...



"""Base classes for retrieval."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any, cast

from typeguard import typechecked


@typechecked
class RetrievalBase:
    """Base Retrieval class."""

    content: Any

    def __init__(self, sources: Any) -> None:
        self.sources = sources

    @abstractmethod
    def get(self, query: str = '') -> Any:
        """Get the data from the sources."""
        ...


@typechecked
class StringRet(RetrievalBase):
    """String Retrieval class."""

    def get(self, query: str = '') -> list[str]:
        """Get the data from the sources."""
        return cast(list[str], self.sources)



"""OpenAI Generation Model class for flexible GPT-based text generation."""

from __future__ import annotations

from typing import cast

import openai

from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class OpenAIGen(GenerationBase):
    """OpenAI generation model for text generation."""

    default_model_name = 'gpt-3.5-turbo'

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        self.model = openai.OpenAI(api_key=self.api_key)

    def generate(
        self,
        query: str,
        context: list[str],
    ) -> str:
        """Generate text using OpenAI's API with dynamic model support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        if not self.model:
            raise Exception('The model was not created.')

        response = self.model.chat.completions.create(
            model=self.model_name,
            messages=[{'role': 'user', 'content': input_text}],
            max_tokens=self.output_max_length,
            temperature=self.temperature,
            top_p=0.9,
            frequency_penalty=0.5,
            presence_penalty=0.3,
        )

        return cast(str, response.choices[0].message.content.strip())



"""Augmented package."""

from __future__ import annotations

from rago.augmented.base import AugmentedBase
from rago.augmented.openai import OpenAIAug
from rago.augmented.sentence_transformer import SentenceTransformerAug

__all__ = [
    'AugmentedBase',
    'OpenAIAug',
    'SentenceTransformerAug',
]



"""GeminiGen class for text generation using Google's Gemini model."""

from __future__ import annotations

from typing import cast

import google.generativeai as genai

from typeguard import typechecked

from rago.generation.base import GenerationBase


@typechecked
class GeminiGen(GenerationBase):
    """Gemini generation model for text generation."""

    default_model_name: str = 'gemini-1.5-flash'

    def _setup(self) -> None:
        """Set up the object with the initial parameters."""
        genai.configure(api_key=self.api_key)
        self.model = genai.GenerativeModel(self.model_name)

    def generate(self, query: str, context: list[str]) -> str:
        """Generate text using Gemini model support."""
        input_text = self.prompt_template.format(
            query=query, context=' '.join(context)
        )

        response = self.model.generate_content(input_text)
        return cast(str, response.text.strip())
